/*
 * Written by Solar Designer and placed in the public domain.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/mman.h>

extern char *__progname;

/* glibc 2.3.6 does not provide a prototype for 5-argument mremap(2) yet */
#include <asm/unistd.h>
#define __NR_mremap5 __NR_mremap
static _syscall5(void *, mremap5,
	void *, addr, size_t, old_len, size_t, new_len, unsigned long, flags,
	void *, new_addr);
#define mremap mremap5
/* Get MREMAP_FIXED */
#include <linux/mman.h>

static long pagesize = 0;

static long init_pagesize(void)
{
	if (!pagesize) {
		pagesize = sysconf(_SC_PAGESIZE);
		if (pagesize == -1)
			perror("sysconf(_SC_PAGESIZE)");
	}
	return pagesize;
}

static void do_mmap(unsigned long start)
{
	void *retval;
	int saved_errno;

	if (init_pagesize() < 0) return;

	retval = mmap((void *)start, pagesize, PROT_READ|PROT_WRITE,
		MAP_PRIVATE|MAP_FIXED|MAP_ANONYMOUS, -1, 0);
	saved_errno = errno;

	printf("mmap(%p, %ld, PROT_READ|PROT_WRITE, "
		"MAP_PRIVATE|MAP_FIXED|MAP_ANONYMOUS) = %p",
		(void *)start, pagesize, retval);
	if (retval == MAP_FAILED)
		printf(" (%s)", strerror(saved_errno));
	putchar('\n');
}

static void do_mremap(unsigned long start)
{
	void *retval, *addr;
	int saved_errno;

	if (init_pagesize() < 0) return;

	retval = mmap(NULL, pagesize, PROT_READ|PROT_WRITE,
		MAP_PRIVATE|MAP_ANONYMOUS, -1, 0);
	if (retval == MAP_FAILED) {
		perror("mmap");
		return;
	}
	addr = retval;

	retval = mremap(addr, pagesize, pagesize,
		MREMAP_MAYMOVE|MREMAP_FIXED, (void *)start);
	saved_errno = errno;

	printf("mremap(%p, %ld, %ld, MREMAP_MAYMOVE|MREMAP_FIXED, %p) = %p",
		addr, pagesize, pagesize, (void *)start, retval);
	if (retval == MAP_FAILED)
		printf(" (%s)", strerror(saved_errno));
	putchar('\n');
}

#define USAGE \
"Usage: %s OPTION\n" \
"Page zero mapping tests\n\n" \
"  -m\t\tcall mmap(2) with a fixed start address of 0 (NULL)\n" \
"  -r\t\tcall mremap(2) with a fixed new start address of 0 (NULL)\n" \
"  -M [ADDRESS]\tcall mmap(2) with the specified fixed start address\n" \
"  -R [ADDRESS]\tcall mremap(2) with the specified fixed new start address\n"

static void usage(void)
{
	printf(USAGE, __progname);
	exit(1);
}

int main(int argc, char **argv)
{
	int c;
	long value;
	char *error;

	while ((c = getopt(argc, argv, "mrM:R:")) != -1) {
		switch (c) {
		case 'm':
			do_mmap(0);
			break;

		case 'r':
			do_mremap(0);
			break;

		case 'M':
		case 'R':
			errno = 0;
			value = strtoul(optarg, &error, 0);
			if (errno || !*optarg || *error)
				usage();
			if (c == 'M')
				do_mmap(value);
			else
				do_mremap(value);
			break;

		default:
			usage();
		}
	}

	if (argc < 2 || optind != argc)
		usage();

	return 0;
}
